#!/bin/bash -e

# This script is meant to be sourced.
# 
# Inputs:
# - $1 = AMI or series/grep string to use
# - $2 = PUBLIC_IP to use (optional)
#   
# Outputs (global variables set):
# - PUBLIC_IP = set to PUBLIC_IP or $IP if no PUBLIC_IP given
# - INSTANCE = instance number
# - IP = IP address of instane, private
#
# Methods:
# - die()
# - srexec()
# - rexec()
# - create_instance()
#
# Env Variables:
# - KEY_NAME = ssh key name you want to use (default: landscape)
# - SECURITY_GROUPS = if set, pass to euca-run-instances -g
#

INTERACTIVE=0                                                                  
tty -s && INTERACTIVE=1   
SSH_OPTS=-qn
[ $INTERACTIVE -eq 1 ] && SSH_OPTS=-q
CLOUDRC="/srv/landscape-scripts/cloud/canonistack/novarc"
INSTANCE_TYPE=${INSTANCE_TYPE:-m1.small}

die() {
    STATUS=$?
    if [ $STATUS -eq 0 ]; then
        STATUS=1
    fi
    echo "ERROR: $1"
    exit $STATUS
}

# Given an instance, terminate
terminate() {
    local instance=$1
    echo "* Attempting to terminate: $instance"
    euca-terminate-instances $instance || /bin/true
}


get_images() {
    if [ $AWS -eq 1 ]; then
        get_images_amazon $1
    else
        get_images_canonistack $1
    fi
}

get_images_canonistack() {
    euca-describe-images |grep smoser-cloud-images/ | grep amd64 \
        | awk '{ print $2 "  " $3 }' | grep "$1" | sort 
}

get_images_amazon() {
    AWS_IMAGES=https://cloud-images.ubuntu.com/query/$1/server/released.current.txt
    curl -s $AWS_IMAGES >& /dev/null || return
    curl -s $AWS_IMAGES | grep instance-store | grep $ARCH | grep $REGION \
        | awk '{ print $8,$1,$2,$3,$4,$5,$6,$7 }'
}

# Given an ami or series string, return the ami to use
parse_ami() {
    ( echo $1 | egrep '^ami-' ) && return
    get_images $1 | grep $1 | tail -1 | awk '{ print $1 }'
}

srexec() {
    ssh $SSH_OPTS root@$PUBLIC_IP $*
}
rexec() {
    ssh $SSH_OPTS ubuntu@$PUBLIC_IP $*
}

# Terminate any instance started by this script
terminate_instance() {
    if [ ! -z "$INSTANCE" ]; then
        euca-terminate-instances $INSTANCE || /bin/true
    fi
}

create_instance() {

    AMI=$(parse_ami $1)
    IP=
    PUBLIC_IP=$2
    INIT=$DIR/cloud-init-hash-bang

    if [ "$SECURITY_GROUPS" == "" ]; then
        GROUP_PARAM=
    else
        for g in $SECURITY_GROUPS; do
            GROUP_PARAM="$GROUP_PARAM -g $g"
        done
    fi


    echo "* Starting up instance ($AMI)"
    INSTANCE=$(euca-run-instances $GROUP_PARAM -t $INSTANCE_TYPE \
        -k $KEY_NAME -f $INIT $AMI \
        | grep INSTANCE | awk '{ print $2 }')

    if [ $? -ne 0 ]; then
        echo "ERROR running instnace: $?"
        exit 1
    fi

    if [ "$INSTANCE" == "" ]; then
        die "did not get instance from euca-run-instances"
    fi

    echo "* Wait for $INSTANCE to be 'running'. (20 attempts)"
    COUNT=1
    while /bin/true; do
        if [ $COUNT -gt 20 ]; then
            terminate $INSTANCE
            die "timeout waiting for instance:$INSTANCE to spawn"
        fi
        COUNT=$(expr $COUNT + 1)

        STATUS=$(euca-describe-instances $INSTANCE | grep INSTANCE | awk -F '\t' '{ print $6 }')
        if [ "$STATUS" != "running" ]; then
            sleep 60 
            continue
        fi
        IP=$(euca-describe-instances $INSTANCE | grep INSTANCE  | awk -F '\t' '{ print $4 }')

        # FIXME: Canonistack changed feb 15th 2012 to start returning names
        # instead IPs out of euca-describe-instances, remove once this is
        # fixed.
        IP=$($DIR/get-ip-of-instance $INSTANCE)
        break

    done

    # Associate and use the public IP, if given.
    if [ "$PUBLIC_IP" == "" ]; then
        PUBLIC_IP=$IP
    else
        echo "* Associate Public IP"
        euca-associate-address -i $INSTANCE $PUBLIC_IP
    fi

    echo "* Network Info: $IP $PUBLIC_IP"
    echo "* Sleeping for a minute hoping to avoid race condition"
    sleep 60 

    echo "* Attempting to SSH, timeout after 50 attempts"
    COUNT=1
    while /bin/true; do
        if srexec 'cat /tmp/done > /dev/null 2>&1'; then
            break
        else
            if [ $COUNT -gt 50 ]; then
                terminate $INSTANCE
                die "timeout waiting for ssh to respond on $PUBLIC_IP"
            fi
            COUNT=$(expr $COUNT + 1)
        fi
        sleep 15
    done

}

bootstrap() {
    echo "* Seeding scripts to new instance, in /tmp"
    scp -qpC $DIR/* root@$PUBLIC_IP:/tmp
}

# On ci.landscape, this is how I source credentials...
if [ -e $HOME/.juju/novarc ]; then
    source $HOME/.juju/novarc
fi
[ "$EC2_URL" == "" ] && die "Need EC2_URL and other CLOUD ENV set up!"

AMI=
DIR="$( cd "$( dirname "$0" )" && pwd )"
KEY_NAME=${KEY_NAME:-landscape}
ARCH=i386
REGION=$(echo $EC2_URL | cut -d . -f 2)
if echo $EC2_URL | grep amazonaws; then
    AWS=1
else
    AWS=0
fi

/bin/true
