#!/bin/bash -x

LICURL="http://people.canonical.com/~andreas/54dee2e5296e360111c2af5d43b8c35aa283e65f/license.txt"
FRESH=-1;
CERT_ONLY=-1;

# Use to write out a custom repository - only include the repository URL
LDS_REPOSITORY=${LDS_REPOSITORY:-}

NOSCHEMA=0;
PRIVATE=
CN=
DIR="$( cd "$( dirname "$0" )" && pwd )"
LSCTL="lsctl"
STATUS=0;
LSENV="/opt/canonical/landscape/scripts/landscape-env.sh"
CODENAME=$(cat /etc/lsb-release | grep CODENAME | sed 's/.*=//')


KEY_FILE=/etc/ssl/private/landscape_server.key
CERT_FILE=/etc/ssl/certs/landscape_server.pem
CA_CERT_FILE=/etc/ssl/certs/landscape_server_ca.crt
SOURCES_FILE=/etc/apt/sources.list.d/landscape-devel-team.list


die() {
    STATUS=$?
    echo "ERROR: $1"
    exit $STATUS 
}

error() {
    STATUS=$?
    echo "ERROR: $1"
}

use_custom_repository() {
    [ -z "$LDS_REPOSITORY" ] && return
    cat > $SOURCES_FILE <<-EOMSG
deb $LDS_REPOSITORY ${CODENAME} main
EOMSG
    chmod go+r $SOURCES_FILE
}

show_help() {
    cat <<EOF
Usage: $0 { --update | --fresh | --fresh-no-schema | --certificate } CN
       $0 { --update | --fresh | --fresh-no-schema | --certificate }
       $0 [--help]

CN would be the IP/Hostname that is reachable by clients/web browsers.
You can & should omit this if running this script
on a cloud instance, as it will determine what to do for you.  CN will
be used to seed the CN on the certificate, and the root_url parameter
in the database.  It's useful if you need to recreate the certificate file
after you have already run quickinstall.

TODO: cleanup text after this is working on clouds better.
EOF
}

freshdata() {
    . $LSENV || die "Something wrong with LDS deployment"

    cd /opt/canonical/landscape/

    LANDSCAPE_CONFIG=standalone-schema ./schema \
        --with-computers \
        --with-free-disk-space \
        --with-free-memory-and-swap \
        --with-load-averages \
        --with-temperatures \
        --with-network-traffic \
        --with-active-processes \
        --with-packages \
        --with-package-activities \
        --with-script-activities \
        --with-users-and-groups \
        --with-user-and-group-activities \
        --with-custom-graph \
        --with-scripts \
        --with-root-url https://${CN}/ \
        --with-system-email landscape-devel@lists.canonical.com \
        --with-account-password foo \
        || die "error in LDS deployment, schema script"
}

function install_lic() {
    wget -O /etc/landscape/license.txt $LICURL
}

function apt_get() {
    DEBIAN_FRONTEND=noninteractive apt-get \
        --allow-unauthenticated -y \
        -o Dpkg::Options::='--force-confold' $*;
}

function install_cert() {
    tmpfile=$(mktemp /tmp/XXXXXX) || exit 1
    cat > $tmpfile <<EOF
RANDFILE                = /dev/urandom

[ req ]
default_bits            = 1024
default_keyfile         = privkey.pem
distinguished_name      = req_distinguished_name
prompt                  = no
policy                  = policy_anything
x509_extensions         = v3_ca

[ req_distinguished_name ]
commonName              = $CN

[ v3_ca ]
# Extensions to add to a certificate request
subjectAltName          = @alt_names

[alt_names]
DNS.1   = $PUBLIC
DNS.2   = $PRIVATE
DNS.3   = $FQDN
DNS.4   = $CN

EOF

    openssl req \
        -new -x509 -nodes -days 3650 \
        -config $tmpfile \
        -out $CERT_FILE \
        -keyout $KEY_FILE
    cp -f $CERT_FILE $CA_CERT_FILE
    rm -f $tmpfile

    echo "Installed new certificate:"
    echo "  Certificate: $CERT_FILE"
    echo "  CA Certificate: $CA_CERT_FILE"
    echo "  Private Key: $KEY_FILE"

}

function add_server_alias() {
    # add all the alternative names to a ServerAlias directive in the
    # apache config, so we can actually use them with SNI
    for f in /etc/apache2/sites-enabled/*; do
        if ! grep -q ServerAlias $f; then
            # This adds ServerAlias below ServerName preserving the ServeRName
            # identation.
            sed -i -r --follow-symlinks \
            '/^[[:blank:]]*ServerName/{p;s/^([[:blank:]]*).*/\1ServerAlias $FQDN $PUBLIC $PRIVATE $CN/;}' $f
        fi
    done
}

function install_lds() {
    apt_get update;
    if [ $FRESH -eq 1 ]; then
        apt_get install \
            landscape-server \
            landscape-server-quickstart
    else
        # upgrade case
        apt_get dist-upgrade
    fi
}

teardown() {
    $LSCTL stop
    for n in $(sudo -u postgres psql -l|grep standalone|awk '{print $1}'); do 
        sudo -u postgres dropdb $n
    done

    sudo -u postgres psql -c "DELETE FROM pg_authid WHERE rolname IN ('landscape_superuser','landscape_maintenance','landscape');"

    apt_get purge landscape-server landscape-server-quickstart

    rm -fr /etc/landscape/service.conf /etc/landscape/server.conf
    rm -f /etc/init.d/landscape-{appserver,pingserver,async-frontend,job-handler,api,combo-loader,msgserver}

    a2dissite $PRIVATE
    a2dissite $FQDN
    a2dissite $PUBLIC
    a2dissite $CN
    rm -f /etc/apache2/sites-available/$PRIVATE
    rm -f /etc/apache2/sites-available/$FQDN
    rm -f /etc/apache2/sites-available/$PUBLIC
    rm -f /etc/apache2/sites-available/$CN
}

# If running in a cloud instance, get hostnames from the metadata url
get_cloud_network() {
    API_VERSION="2008-02-01"
    METADATA_URL="http://169.254.169.254/$API_VERSION/meta-data"

    # short circuit if not on a cloud
    wget -q -O - ${METADATA_URL}/public-hostname || return

    # if on a cloud, set ips/names correctly.
    FQDN=$(curl -s ${METADATA_URL}/public-hostname)
    # re-enable once canonistack supports /dpb/17/feb/2012
    #PUBLIC=$(curl -s ${METADATA_URL}/public-ipv4)
    PRIVATE=$(curl -s ${METADATA_URL}/local-ipv4)

    echo network data: $FQDN $PUBLIC $PRIVATE
}


get_options() {
    while true; do
        case "$1" in
          -h|--help|-\?) show_help; exit 0;;
          -u|--update) FRESH=0; break;;
          -f|--fresh) FRESH=1; break;;
          -n|--fresh-no-schema) FRESH=1; NOSCHEMA=1; break;;
          --certificate) CERT_ONLY=1; break;;
          --) shift; break;;
          -*) echo "invalid option: $1" 1>&2; show_help; exit 1;;
          *)  break;;
        esac
    done
    CN=$2
    if [ $FRESH -eq -1 -a $CERT_ONLY -eq -1 ]; then
        show_help
        exit 1
    fi
}

## MAIN

get_options $@
get_cloud_network

[ "$PUBLIC" == "" ] && die "no PUBLIC ip set"
[ "$PRIVATE" == "" ] && PRIVATE=$PUBLIC
[ "$FQDN" == "" ] && FQDN=$PUBLIC

use_custom_repository

if [ $CERT_ONLY -eq 1 ]; then
    [ "$CN" == "" ] && CN=$FQDN
    install_cert
    exit $STATUS
else
    [ "$CN" == "" ] && CN=$PUBLIC
fi

if [ $FRESH -eq 1 ]; then
    teardown
fi

install_cert
install_lic
install_lds
add_server_alias
service apache2 restart

# use a different lsctl if we are using an older version of
# landscape
dpkg -l landscape-server | grep '11\.07\.201108'
if [ $? -eq 0 ]; then
    LSCTL="$DIR/lsctl.11.07"
fi


$LSCTL status || error "Not all services running."

if [ $FRESH -eq 1 ]; then
    if [ $NOSCHEMA -eq 0 ]; then
        $LSCTL stop || error "Not all services running."
        freshdata
        $LSCTL start || error "Not all services started."
    fi
fi


exit $STATUS
