#!/bin/bash

DIR="$( cd "$( dirname "$0" )" && pwd )"
TOLERANCE=2
DEAD_DB=/tmp/slave-garbage-dead
TERMINATED_DB=/tmp/slave-garbage-terminated
HIGHLANDER_DB=/tmp/slave-garbage-highlander
OPT=$1

# $1 = INSTANCE
# $2 = Optional message suffix
terminate() {
    if [ "$OPT" == "--cleanup" ]; then
        echo -n " - Terminating ${2}"
        highlander_check $1
        euca-terminate-instances $INSTANCE > /dev/null
    else
        echo -n " - Would terminate ${2}"
    fi
}

# Make sure I can contact to avoid long delays on all commands
ping_slave() {
    ssh ubuntu@$1 -qn "/bin/true" && return
    echo $2 >> $DEAD_DB
    #dead_check
    return 1
}

# Given an instance, check how many times it's been reported dead
dead_check() {
    local num=$(grep $INSTANCE $DEAD_DB | wc -l)
    if [ $num -gt 5 ]; then
        terminate $INSTANCE ", DEAD (n=$num)"
    else
        echo -n " - not alive, (n=$num)"
    fi
}

# Make sure I keep track of instances that can't be terminated
# Just keep track of terminated instances in a file.  If I get to a 
# tolerance count in the file, mark it as a true highlander
highlander_check() {
    local id=$1
    echo $id >> $TERMINATED_DB

    local num=$(grep $id $TERMINATED_DB | wc -l)
    if [ $num -gt 10 ]; then
        echo $id >> $HIGHLANDER_DB
    fi
}


# Make sure instance is indeed a slave
is_slave() {
    ssh ubuntu@$1 -qn "test -d /var/jenkins" && return 0
    echo -n " - no /var/jenkins dir"
    return 1
}

# Return:
#        0 = slave disconnected
#       !0 = slave still connected
slave_disconnected() {
    status=$(ssh ubuntu@$1 -qn "ps -ef | grep jenkins | grep s[l]ave.jar")

    if [ $? -eq 0 ]; then
        echo -n " - slave still connected"
        return 1
    else
        return 0
    fi
}

# $1 = IP to check
# Returns 0 if address is a private ip, 1 otherwise
private_ip() {
    local ip=$1
    [ -z $ip ] && return 0
    echo $ip | grep '^91.' > /dev/null || return 0
    echo -n " - skip public IPs ($ip)"
    return 1
}

# $1 = offset since start of instance in hours
# Returns 0 if >= tolerance, 1 otherwise
old_instance() {
    local offset=$1
    [ $offset -ge $TOLERANCE ] && return 0
    echo -n " - Young instance, skipping"
    return 1
}

# Get a list of instances, Expected return right now is
# INSTANCE STATE CREATED_AT IP 
running_instances() {
    $DIR/list-instances | awk -F "\t" '{ print $1,$3,$4,$2 }'
}

zombie_state() {
    [ "$1" == "pending" ] && return 0
    [ "$1" == "terminating" ] && return 0
    [ "$1" == "terminated" ] && return 0
    [ "$1" == "error" ] && return 0
    return 1
}

if [ "$1" == "" ]; then
    cat <<-EOF
Usage:
    $0 --cleanup
    $0 --report

Running Instances:
------------------
EOF
    running_instances
    exit 1
fi

running_instances | while read line; do
    read INSTANCE STATE CREATED_AT IP <<<$(echo $line)
    echo

    HOURS_OFFSET=$(echo $CREATED_AT | sed 's/:.*//')
    echo -n "$INSTANCE $IP $STATE $CREATED_AT (${HOURS_OFFSET}h)"

    # Just for sanity, don't touch young things, or things with public ips
    old_instance $HOURS_OFFSET || continue
    private_ip $IP $INSTANCE || continue

    # Short circuit: pending (e.g.) after tolerance?  kill it.
    if zombie_state $STATE; then
        terminate $INSTANCE Zombie
        continue; 
    fi

    # Sentials
    ping_slave $IP $INSTANCE || continue;
    is_slave $IP $INSTANCE || continue;
    slave_disconnected $IP $INSTANCE || continue;

    # Reap
    terminate $INSTANCE
done

echo
