#!/usr/bin/python

import logging
import os
import optparse
from subprocess import check_output, STDOUT, CalledProcessError, PIPE, Popen
import sys

from shutil import rmtree

charms = {
    "landscape": 
        {"stable": "lp:~landscape/landscape-charm/stable",
         "trunk": "lp:~landscape/landscape-charm/trunk"},
    "postgresql":
        {"stable": "lp:~landscape/landscape-charm/postgresql",
         "trunk": "lp:charms/precise/postgresql"},
    "rabbitmq-server":
        {"stable": "lp:~landscape/landscape-charm/rabbitmq-server",
         "trunk": "lp:charms/precise/rabbitmq-server"},
    "apache2":
        {"stable": "lp:~landscape/landscape-charm/apache2",
         "trunk": "lp:charms/precise/apache2"},
    "haproxy":
        {"stable": "lp:~landscape/landscape-charm/haproxy",
         "trunk": "lp:charms/precise/haproxy"},
    "juju-gui":
        {"stable": "lp:~landscape/landscape-charm/juju-gui",
         "trunk": "lp:charms/precise/juju-gui"}
}

def checkout(branch, directory):
    logging.info("Checking out %s into %s" % (branch,directory))
    cmd = ["bzr", "co", "--lightweight", branch, directory]
    try:
        check_output(cmd, stderr=STDOUT)
    except CalledProcessError as e:
        logging.error("checkout failed, code %d, output: %s" %
                     (e.returncode,e.output))
        sys.exit(1)

def merge(branch, directory):
    logging.info("Merging %s into %s" % (branch, directory))
    try:
        output = check_output(["bzr", "merge", branch], cwd=directory,
                              stderr=STDOUT)
    except CalledProcessError as e:
        logging.error("merge failed, code %d, output: %s" %
                     (e.returncode, e.output))
        sys.exit(1)
    else:
        if output.strip() == "Nothing to do.":
            return False
        return True

def commit(directory, message):
    logging.info("Committing %s" % directory)
    try:
        check_output(["bzr", "commit", "-m %s" % message], cwd=directory,
                     stderr=STDOUT)
    except CalledProcessError as e:
        logging.error("commit failed, code %d, output: %s" %
                     (e.returncode, e.output))
        return False
    else:
        return True

def init_repo(directory):
    logging.info("Initializing bzr shared repository at %s" % directory)
    try:
        check_output(
            ["bzr", "init-repo", directory], stderr=STDOUT)
    except CalledProcessError as e:
        logging.error("init-repo failed, code %d, output: %s" %
                     (e.returncode, e.output))
        sys.exit(1)

def checkout_charms(release, path):
    for charm in charms.keys():
        charm_directory = path + "/" + charm
        # for some reason, bzr up fails on a checked out landscape charm directory,
        # so let's start fresh everytime
        rmtree(charm_directory, ignore_errors=True)
        checkout(charms[charm][release], charm_directory)

def merge_charms(path):
    for charm in charms.keys():
        charm_directory = path + "/" + charm
        changed = merge(charms[charm]["trunk"], charm_directory)
        charms[charm]["changed"] = changed
        if changed:
            logging.info("No changes for %s" % charm)
        else:
            logging.info("Merged changes for %s" % charm)

def commit_charms(path, message):
    for charm in charms.keys():
        if not charms[charm]["changed"]:
            logging.info("Skipping charm %s, no changes to commit" % charm)
            continue
        charm_directory = path + "/" + charm
        if not commit(charm_directory, message):
            return False
    return True

def deploy(path, target, repository_url, license_file):
    cmd = ["./deploy-charm-test", path, target]
    prefix = "%s/landscape/config" % path
    with open("%s/repo-file" % prefix, "w") as repository_file:
        repository_file.write(repository_url + "\n")
    license = open(license_file, "r").read()
    with open("%s/license-file" % prefix, "w") as lf:
        lf.write(license)
    logging.info("Calling deployment script: %s" % cmd)
    try:
        process = Popen(cmd, bufsize=1, stdout=PIPE, stderr=STDOUT)
    except CalledProcessError as e:
        logging.error(e.output)
        logging.error("Deployment failed")
        return False
    else:
        while process.poll() is None:
            logging.info(process.stdout.readline().strip())
        return process.returncode == 0
    
def get_option_parser():
    usage = "%prog [options]"
    description = ("Script to test-deploy the landscape charm, optionally "
                   "merging branches if the deployment worked.")
    parser = optparse.OptionParser(usage=usage, description=description)
    parser.add_option("-l", "--license-file", dest="license_file",
        help="LDS license file to use")
    parser.add_option("-r", "--repository-url", dest="repository_url",
        help="URL of the LDS repository to use. Use this to also "
        "implicitly select which LDS version you want.")
    parser.add_option("-d", "--deployment-type", dest="deployment_type",
        default="trunk", help="Deploy trunk or stable versions "
        "of the charms [default: %default]")
    parser.add_option("-m", "--merge", dest="merge", action="store_true",
        default=False, help="Merge (but not commit) trunk into stable "
        "before deploying [default: %default]")
    parser.add_option("-c", "--commit", dest="commit", default=False,
        action="store_true", help="When merging, commit if the deployment "
        "succeeds [default: %default]")
    parser.add_option("--root-directory", dest="root_directory",
        help="Where to branch the charms into. Will be created if it does "
        "not exist.")
    parser.add_option("--deployment-target", dest="deployment_target",
        default="landscape", help="The deployment target from the "
        "juju-deployer configuration file that should be used [default: "
        "%default]")
    parser.add_option("--commit-message", dest="commit_message",
        default="Merged from upstream", help="Commit message to use when "
        "the commit option is used.")
    return parser    


logging.basicConfig(level=logging.INFO,
    format='%(asctime)s %(levelname)s %(message)s')

parser = get_option_parser()
options, args = parser.parse_args()

if len(args):
    parser.error("No extra arguments.")

if not options.license_file:
    parser.error("Need license_file")

if not options.repository_url:
    parser.error("Need repository_url")

if not options.root_directory:
    parser.error("Need a root_directory")

if options.deployment_type not in ["trunk", "stable"]:
    parser.error("Deployment type must be either trunk or stable.")

if (options.merge or options.commit) and options.deployment_type == "trunk":
    parser.error("Merge or commit option can only be used with the stable "
                 "deployment option")

if options.commit and not options.merge:
    parser.error("Commit cannot be used without the merge option")

if not os.path.isdir(options.root_directory):
    os.makedirs(options.root_directory)
    init_repo(options.root_directory)

logging.info("Checking out %s charms" % options.deployment_type)
checkout_charms(options.deployment_type, options.root_directory)

if options.merge:
    logging.info("Merging trunk into stable")
    merge_charms(options.root_directory)

result = deploy(options.root_directory, options.deployment_target,
                options.repository_url, options.license_file)

if not result:
    logging.info("Deployment failed.")
    sys.exit(1)
else:
    logging.info("Deployment succeeded.")

if options.commit and result:
    logging.info("Commit requested.")
    result = commit_charms(options.root_directory, options.commit_message)
    if not result:
        logging.info("Commit failed.")
        sys.exit(1)
    else:
        logging.info("Commit succeeded.")

sys.exit(0)
