#!/usr/bin/python

import sys
import os
import logging
import optparse
from subprocess import Popen, PIPE, STDOUT, check_call, CalledProcessError


class UncommittedChanges(Exception):
    pass


class InvalidMergeTarget(Exception):
    pass


class InvalidRepository(Exception):
    pass


class CommandError(Exception):

    def __init__(self, command, path, error_code, output):
        self.command = command
        self.error_code = error_code
        self.output = output
        self.path = path

    def __str__(self):
        output = self.output.split("\n")[-5:]
        info_text = "Command: %s\nPath: %s\nerror code: %u\nOutput:\n" % (
                    " ".join(self.command), self.path, self.error_code)
        return info_text + "\n(...)\n" + "\n".join(output)


class TestError(Exception):
    pass


class Repository(object):

    # XXX - need support for running more than one command
    # for each action
    revno_cmd = ["bzr", "revno"]
    test_cmd = ["make", "test"]
    clean_cmd = ["make", "clean"]
    merge_cmd = ["bzr", "merge"]
    commit_cmd = ["bzr", "commit"]
    status_cmd = ["bzr", "diff"]
    revert_cmd = ["bzr", "revert"]
    build_cmd = ["make"]
    # assumes a bound repository
    update_cmd = ["bzr", "update"]
    tag_cmd = ["bzr", "tag", "--force"]
    test_ok = False
    CHANGED_STATUS = 1
    NOCHANGE_STATUS = 2
    # Should always be overwritten
    commit_message = "No message, please fix"

    def __init__(self, path, test_cmd=None, clean_cmd=None, revision=None):
        self.path = path
        self.name = os.path.basename(os.path.dirname(path))
        self.revision = revision
        if not os.path.isdir(self.path):
            raise IOError("Not a directory: %s\n" % self.path)
        # XXX - also check if self.path points to a valid bzr repository
        if test_cmd:
            self.test_cmd = test_cmd
        if clean_cmd:
            self.clean_cmd = clean_cmd

    def get_path(self):
        return self.path

    def get_status(self):
        try:
            cmd = Popen(self.status_cmd, cwd=self.get_path(),
                        stdout=open("/dev/null"), stderr=STDOUT)
        except OSError:
            output = "Error, command \"%s\" not found" % " ".join(cmd)
            raise CommandError(cmd, None, output)
        output = cmd.communicate()[0]
        errcode = cmd.returncode
        self.status_output = output
        if errcode in [1, 2]:
            return self.CHANGED_STATUS
        elif errcode == 0:
            return self.NOCHANGE_STATUS
        else:
            raise CommandError(cmd, self.get_path(), errcode, output)
        # not reached
        return False

    def _get_command_output(self, cmd, cwd=None, env=os.environ.copy()):
        if cwd is None:
            cwd = self.get_path()
        env["PWD"] = cwd
        popen_kwargs = {'stdout': PIPE, 'stderr': STDOUT}
        logging.debug("Command: %s" % cmd)
        logging.debug("CWD: %s" % cwd)
        logging.debug("Environment: %s" % env)

        try:
            command = Popen(cmd, cwd=cwd, env=env, **popen_kwargs)
        except OSError, e:
            output = ("Error, command \"%s\" not found (%s)" %
                     (" ".join(cmd), e))
            raise CommandError(cmd, cwd, None, output)
        output = command.communicate()[0]
        logging.debug("Command Output: %s" % output)
        errcode = command.returncode
        if errcode != 0:
            raise CommandError(cmd, cwd, errcode, output)
        return output

    def tag(self, tag):
        logging.debug("Tagging repository %s with %s" %
                      (self.get_path(), tag))
        result = self._get_command_output(self.tag_cmd + [tag])
        logging.debug("%s" % result)
        logging.debug("Tagging of repository %s with %s finished" %
                      (self.get_path(), tag))
        return True

    def changed(self):
        if self.get_status() == self.CHANGED_STATUS:
            return True
        else:
            return False

    def get_revno(self):
        result = self._get_command_output(self.revno_cmd)
        return int(result.strip())

    def test(self):
        logging.debug("Starting test of repository %s" % self.get_path())
        try:
            self.test_output = self._get_command_output(self.test_cmd)
        except CommandError, e:
            self.test_output = e.output
            raise TestError()
        self.test_ok = True
        logging.debug("Test of repository %s finished" % self.get_path())
        return True

    def clean(self):
        logging.debug("Starting clean of repository %s" % self.get_path())
        clean_tree_cmd = ["bzr", "clean-tree", "--force", "--detritus"]
        for command in [clean_tree_cmd, self.clean_cmd]:
            try:
                self.clean_output = self._get_command_output(command)
            except CommandError, e:
                self.clean_output = e.output
                raise
        logging.debug("Clean output:\n%s" % self.clean_output)
        logging.debug("Clean of repository %s finished" % self.get_path())
        return True

    def update(self):
        logging.debug("Starting update of repository %s" % self.get_path())
        try:
            self.update_output = self._get_command_output(self.update_cmd)
        except CommandError, e:
            self.update_output = e.output
            raise
        logging.debug("Update output:\n%s" % self.update_output)
        logging.debug("Update of repository %s finished" % self.get_path())
        return True

    def merge(self, source_repository):
        logging.debug("Starting merge of repository %s with %s" %
                      (source_repository.get_path(), self.get_path()))
        command = list(self.merge_cmd)
        if self.revision is not None:
            logging.info("Merging @revision: %s" % self.revision)
        command.append(source_repository.get_path())
        try:
            self.merge_output = self._get_command_output(command)
        except CommandError, e:
            self.merge_output = e.output
            raise
        logging.debug("Merge output:\n%s" % self.merge_output)
        logging.debug("Merge of repository %s finished" % self.get_path())
        return True

    def build(self):
        logging.debug("Starting build of repository %s" % self.get_path())
        try:
            self.build_output = self._get_command_output(self.build_cmd)
        except CommandError, e:
            self.build_output = e.output
            raise
        logging.debug("Build output:\n%s" % self.build_output)
        logging.debug("Build of repository %s finished" % self.get_path())
        return True

    def revert(self):
        logging.debug("Starting revert of repository %s" % self.get_path())
        try:
            self.revert_output = self._get_command_output(self.revert_cmd)
        except CommandError, e:
            self.revert_output = e.output
            raise
        logging.debug("Revert output:\n%s" % self.revert_output)
        logging.debug("Revert of repository %s finished" % self.get_path())
        return True

    def prep_for_merge(self):
        logging.debug("Preparing repository %s for merge" % self.get_path())
        self.clean()
        self.revert()
        if self.changed():
            raise UncommittedChanges(
                "Repository %s has uncommitted changes" % self.get_path())
        self.update()
        logging.debug("Repository %s is ready for merge" % self.get_path())
        return True

    def commit(self):
        logging.debug("Starting commit of repository %s" % self.get_path())
        command = self.commit_cmd + ["-m", self.commit_message]
        try:
            self.commit_output = self._get_command_output(command)
        except CommandError, e:
            self.commit_output = e.output
            raise
        logging.debug("Commit output:\n%s" % self.commit_output)
        logging.debug("Commit of repository %s is finished" % self.get_path())
        return True


class txawsRepository(Repository):
    test_cmd = ["trial", "txaws"]
    clean_cmd = ["/bin/true"]
    build_cmd = ["/bin/true"]


class zope3Repository(Repository):
    pass


class lazrjsRepository(Repository):
    test_cmd = None
    build_cmd = ["/bin/true"]


class landscapeClientRepository(Repository):
    test_cmd = ["make", "check"]


class convoyRepository(Repository):
    test_cmd = ["python", "setup.py", "test"]
    build_cmd = ["/bin/true"]
    clean_cmd = ["python", "setup.py", "clean"]


class landscapeRepository(Repository):
    test_cmd = ["bin/test", "canonical"]
    build_cmd = ["make", "build"]
    patch_cmd = ["make", "patch"]
    copydb_cmd = ["lsd", "copy-databases"]

    def __init__(self, path, test_cmd=None, clean_cmd=None, revision=None,
                 no_db_test=False):
        super(landscapeRepository, self).__init__(
            path, test_cmd, clean_cmd, revision)
        self.no_db_test = no_db_test

    # overridden because we need the env var LANDSCAPE_CONFIG
    def build(self):
        logging.debug("Starting build of repository %s" % self.get_path())
        env = os.environ.copy()
        target = os.path.basename(self.get_path().rstrip("/"))
        env["LANDSCAPE_CONFIG"] = target
        try:
            self.build_output = self._get_command_output(
                self.build_cmd, env=env)
        except CommandError, e:
            self.build_output = e.output
            raise
        logging.debug("Build output:\n%s" % self.build_output)
        logging.debug("Build of repository %s finished" % self.get_path())
        return True

    def copydb(self):
        logging.debug("Copying databases")
        try:
            self.copydb_output = self._get_command_output(self.copydb_cmd)
        except CommandError, e:
            self.copydb_output = e.output
            raise
        logging.debug("copydb output:\n%s" % self.copydb_output)
        logging.debug("Database copy finished")
        return True

    def patch(self):
        logging.debug("Running patch command")
        try:
            self.patch_output = self._get_command_output(self.patch_cmd)
        except CommandError, e:
            self.patch_output = e.output
            raise
        logging.debug("Patch output:\n%s" % self.patch_output)
        logging.debug("Patch application finished")
        return True

    def test(self):
        self.test_output = ""
        if self.no_db_test:
            logging.info("... ... Skipping DB test as requested")
        else:
            self.copydb()
            try:
                self.patch()
            except CommandError:
                self.test_output = self.patch_output
                raise TestError()
            self.test_output = self.patch_output
        logging.debug("Starting test of repository %s" % self.get_path())
        try:
            self.test_output += self._get_command_output(self.test_cmd)
        except CommandError, e:
            self.test_output += e.output
            raise TestError()
        self.test_ok = True
        logging.debug("Test output:\n%s" % self.test_output)
        logging.debug("Test of repository %s finished" % self.get_path())
        return True


class pyamqplibRepository(Repository):
    test_cmd = ["tests/client_0_8/run_all.py"]
    clean_cmd = ["/bin/true"]
    build_cmd = ["python", "setup.py", "build"]

    # overriden because we need to add PYTHONPATH to the env
    def test(self):
        logging.debug("Starting test of repository %s" % self.get_path())
        env = os.environ.copy()
        env["PYTHONPATH"] = self.get_path() + "/build/lib.linux-" + \
            os.uname()[-1] + "-2.6/"
        try:
            self.test_output = self._get_command_output(
                self.test_cmd, env=env)
        except CommandError, e:
            self.test_output = e.output
            raise TestError()
        self.test_ok = True
        logging.debug("Test output:\n%s" % self.test_output)
        logging.debug("Test of repository %s finished" % self.get_path())
        return True


class stormRepository(Repository):
    test_cmd = ["make", "PYTHON=python", "check"]
    build_cmd = ["make", "PYTHON=python"]

    # temporarily (tm) overriden because the django tests fail, and we
    # don't really care about that in landscape
    # Ter Fev 19 12:08:36 BRT 2013 - andreas
    def test(self):
        logging.debug("Starting test of repository %s" % self.get_path())
        logging.warn("Removing django tests, they fail")
        os.system("rm -rf %s/tests/django" % self.get_path())
        try:
            self.test_output = self._get_command_output(self.test_cmd)
        except CommandError, e:
            self.test_output = e.output
            raise TestError()
        self.test_ok = True
        logging.debug("Test of repository %s finished" % self.get_path())
        return True


class LandscapeMerge(object):

    # can't use a dict because I rely on the order
    repository_name_class = [
        ("storm", stormRepository),
        ("txaws", txawsRepository),
        ("convoy", convoyRepository),
        ("landscape-client", landscapeClientRepository),
        ("landscape", landscapeRepository)
    ]
    # but we do use a dict for lookup
    repository_name_class_dict = dict(repository_name_class)

    TEST_OUTPUT_LINES = 90

    def __init__(self, target, repository_root,
                 repository_list=[name[0] for name in repository_name_class],
                 revert_only=False, prepare_only=False, commit=False, tag=None,
                 no_db_test=False, skip_all_tests=False,
                 landscape_revision=None, landscape_source=None):

        if target not in ["production", "staging"]:
            raise InvalidMergeTarget(
                "Merge target needs to be one of \"production\" or"
                "staging\", not \"%s\"" % target)
        # target = staging: merge is trunk -> staging
        # target = production: merge is staging -> production
        self.skip_all_tests = skip_all_tests
        target = target
        self.tag = tag

        if target == "staging":
            source = "trunk"
        elif target == "production":
            source = "staging"

        self.revert_only = revert_only
        self.prepare_only = prepare_only
        self.commit = commit
        self.no_db_test = no_db_test
        self.landscape_revision = landscape_revision
        # build list of repositories
        # order is important
        # [ (target1,source1), (target2,source2), ...]
        self.repository_root = repository_root.rstrip("/")
        os.environ["LANDSCAPE_SOURCE_DEPENDENCIES"] = self.repository_root
        self.repositories = []

        for name in repository_list:
            repository_source = source
            if name == "landscape" and landscape_source is not None:
                repository_source = landscape_source

            target_path = self.repository_root + "/" + name + "/" + target
            source_path = self.repository_root + "/" + name + "/" + repository_source
            try:
                target_repo = self._create_repository_object(name, target_path)
                source_repo = self._create_repository_object(name, source_path)
                # Commit message is target dependent, add it there.
                target_repo.commit_message = "Merging from %s for %s deployment" % (
                    repository_source, target)
                if self.tag:
                    target_repo.commit_message += " of %s" % self.tag
            except InvalidRepository as error:
                print error.args[0]
                sys.exit(1)
            self.repositories.append((target_repo, source_repo))

    def _create_repository_object(self, name, path):
        try:
            repo = self.repository_name_class_dict[name]
        except KeyError:
            raise InvalidRepository(
                "Error, don't know how to handle repository \"%s\"" % path)
        if name == "landscape":
            return repo(
                path, no_db_test=self.no_db_test,
                revision=self.landscape_revision)
        else:
            return repo(path)

    def revert(self):
        logging.info("Reverting target and source repositories %s" %
                     ", ".join([x[0].name for x in self.repositories]))
        for target, source in self.repositories:
            target.clean()
            target.revert()
            source.clean()
            source.revert()

    def prepare(self):
        logging.info("Preparing for merge target and source repositories "
                     "%s" % ", ".join([x[0].name for x in self.repositories]))
        for target, source in self.repositories:
            target.prep_for_merge()
            source.prep_for_merge()

    def apply_tag(self):
        """Apply the tag to all repositories we are handling."""
        for repository in self.repositories:
            logging.info("Tagging %s with %s" % (repository[0].name,
                                                 self.tag))
            try:
                repository[0].tag(self.tag)
            except CommandError:
                logging.warning("Tagging failed!")

    def run(self):
        return_code = 0
        if self.revert_only:
            self.revert()
            return True
        if self.prepare_only:
            self.prepare()
            return True
        logging.info("Preparing repositories for merge and merging...")
        changed_repositories = []
        for target, source in self.repositories:
            target.prep_for_merge()
            source.prep_for_merge()
            target.merge(source)
            if target.changed():
                changed_repositories.append(target)
        logging.info("Building...")
        for target, source in self.repositories:
            target.build()
        logging.info("Testing...")
        skipped = 0
        for target, source in self.repositories:
            logging.info("... %s" % target.path)
            try:
                if target.test_cmd is not None and not self.skip_all_tests:
                    target.test()
                else:
                    skipped += 1
                    logging.warning("No test defined for %s, skipping" %
                                    target.get_path())
            except TestError:
                logging.error("Test for %s failed" % target.get_path())
        logging.info("Finished.")
        if not self.skip_all_tests:
            failed_repositories = [
                repository[0] for repository in self.repositories
                if repository[0].test_cmd and not repository[0].test_ok]
        else:
            failed_repositories = []
        if len(failed_repositories) == 0:
            print "No test failures (%u skipped)" % skipped
        else:
            return_code = 1
            print "The following repositories had test failures:"
            for repository in failed_repositories:
                print "Path: ", repository.get_path()
                print "Last %u lines of test output:" % self.TEST_OUTPUT_LINES
                output = repository.test_output.split("\n")
                print "\n".join(output[-self.TEST_OUTPUT_LINES:])
            print

        if len(changed_repositories) == 0:
            print "No repositories had changes after the merge."
            if self.tag and self.commit and len(failed_repositories) == 0:
                self.apply_tag()
        else:
            print "The following repositories changed after the merge:"
            print "\n".join([repository.get_path()
                            for repository in changed_repositories])
            print
            if len(failed_repositories) == 0 and self.commit:
                print "No tests failed and commit requested, so let's commit!"
                for repository in changed_repositories:
                    logging.info("Committing %s..." % repository.name)
                    repository.commit()
                if self.tag:
                    self.apply_tag()
                print "Done!"
                print "Revision numbers:"
                for repository in self.repositories:
                    print "%s: r%d" % (repository[0].get_path(),
                                       repository[0].get_revno()),
                    if repository[0] not in changed_repositories:
                        print "(no change)"
                    else:
                        print
        return return_code


def get_option_parser():
    usage = "%prog [options] <production|staging>"
    description = (
        "This tool will merge the repositories in preparation for a "
        "production or staging release. It will run a clean, revert, update, "
        "merge, build, test and commit (if so desired). For the Landscape "
        "repository, it will also load the sanitized production database "
        "and run a schema update as part of the test.")
    parser = optparse.OptionParser(usage=usage, description=description)
    parser.add_option(
        "-p", "--prepare-for-merge-only", dest="prepare_only", default=False,
        action="store_true", help="Only prepare for merge (revert, up, clean)"
        " [default: %default]")
    parser.add_option(
        "-r", "--repositories", help="Only work with the supplied repository "
        "list, comma separated, in that order [default: %default].",
        default=", ".join([name[0] for name in
                          LandscapeMerge.repository_name_class]))
    parser.add_option(
        "--revert", help="Only do a revert, leaving things as if they were "
        "untouched [default: %default].", action="store_true",
        dest="revert_only", default=False)
    parser.add_option(
        "-c", "--commit", help="Commit at the end if everything worked "
        "[default: %default]", action="store_true", dest="commit",
        default=False)
    parser.add_option(
        "-t", "--tag", help="Tag each repository with TAG after commit "
        "[default: %default]", dest="tag", default=None)
    parser.add_option(
        "--repository-root", help="Path to where your Landscape "
        "repositories are. We expect to find each repo under "
        "<root>/<reponame>/<type>, where <type> is one of production/staging"
        "/trunk [default: %default]", default="$HOME/canonical/source")
    parser.add_option(
        "--no-db-test", help="Don't copy production sanitized DB nor apply "
        "the patch [default: %default]", default=False, action="store_true")
    parser.add_option(
        "-d", help="Enable debug mode [default: %default]", dest="debug",
        action="store_true", default=False)
    parser.add_option(
        "--skip-all-tests", help="Skip all tests. DANGEROUS! "
        "[default: %default]", default=False, dest="skip_all_tests",
        action="store_true")
    parser.add_option(
        "--landscape-revision", help="Revspec to add to landscape merge cmd.")
    parser.add_option(
        "--source-branch-name", help="Source branch in merge operation. "
        "Branch name only, user and project is hardcoded.")

    return parser


def main():
    parser = get_option_parser()
    options, args = parser.parse_args()

    if len(args) != 1:
        parser.error("Missing release, either production or staging")

    if args[0] not in ["production", "staging"]:
        parser.error("Release must be either production or staging")

    release = args[0]

    if not options.commit and options.tag:
        parser.error("Tagging is only allowed if commit is also specified")

    repository_root = os.path.expandvars(options.repository_root)

    if "landscape" in options.repositories.split(", "):
        # landscape's clean command eventually calls sudo, which will stop
        # the whole process and ask for the password. Since this happens
        # quite a while after we start, the developer might have left the
        # console already and won't see it, only to return several minutes
        # later and find out the merge process has stopped. So we trigger
        # a sudo password question here just so it's cached.
        print("If requested, please type in your password so sudo can cache "
              "it while the merge runs.")
        try:
            check_call(["sudo", "/bin/true"])
        except CalledProcessError as error:
            print("Wrong password or some other error, aborting.")
            sys.exit(error.returncode)
    if options.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    repository_list = [repository.strip() for repository in
                       options.repositories.split(",")]
    logging.basicConfig(
        level=level, format='%(asctime)s %(levelname)s %(message)s')
    landscape_merge = LandscapeMerge(
        release, repository_root=repository_root,
        repository_list=repository_list,
        revert_only=options.revert_only, commit=options.commit,
        prepare_only=options.prepare_only,
        tag=options.tag, no_db_test=options.no_db_test,
        skip_all_tests=options.skip_all_tests,
        landscape_revision=options.landscape_revision,
        landscape_source=options.source_branch_name)
    logging.info("Merging landscape for %s release" % release)
    return landscape_merge.run()

if __name__ == "__main__":
    sys.exit(main())
